from typing import Callable
from typing import TypeVar, Generic

K = TypeVar('K')
V = TypeVar('V')


class Dict(Generic[K, V], dict[K, V]):
    def keep(self, predicate: Callable[[K, V], bool]) -> "Dict[K, V]":
        return Dict({k: v for k, v in self.items() if predicate(k, v)})

    def omit(self, predicate: Callable[[K, V], bool]) -> "Dict[K, V]":
        return Dict({k: v for k, v in self.items() if not predicate(k, v)})



from typing import Union

from pydantic import BaseModel


class PropertyBaseModel(BaseModel):
    """
    Workaround for serializing properties with pydantic until
    https://github.com/samuelcolvin/pydantic/issues/935
    is solved
    """
    @classmethod
    def get_properties(cls):
        return [prop for prop in dir(cls) if isinstance(getattr(cls, prop), property) and prop not in ("__values__", "fields")]

    def dict(
            self,
            *,
            include: Union['AbstractSetIntStr', 'MappingIntStrAny'] = None,
            exclude: Union['AbstractSetIntStr', 'MappingIntStrAny'] = None,
            by_alias: bool = False,
            skip_defaults: bool = None,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
    ) -> 'DictStrAny':
        attribs = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none
        )
        props = self.get_properties()
        # Include and exclude properties
        if include:
            props = [prop for prop in props if prop in include]
        if exclude:
            props = [prop for prop in props if prop not in exclude]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})

        return attribs



import enum

from easy_pysy.core import logging
from easy_pysy.core.event import Event, emit
from easy_pysy.core.signal import sigint_callback
from easy_pysy.utils.common import require


class AppState(enum.Enum):
    STOPPED = 1
    STARTING = 2
    STARTED = 3
    STOPPING = 4


class EzContext:
    state = AppState.STOPPED


context = EzContext()


def start():
    require(context.state == AppState.STOPPED, f"Can't start application, current state: {context.state}]")

    logging.info('Starting')
    context.state = AppState.STARTING
    emit(AppStarting())

    logging.info('Started')
    context.state = AppState.STARTED
    emit(AppStarted())


def stop():
    require(context.state == AppState.STARTED, f"Can't stop application, current state: {context.state}")


    try:
        logging.info(f'Stopping')
        context.state = AppState.STOPPING
        emit(AppStopping())
    except BaseException:
        logging.exception('Error while stopping application ! Some AppStopping subscribers may not have been called')

    logging.info('Stopped')
    context.state = AppState.STOPPED


@sigint_callback
def shutdown(exit_code=0):
    stop()
    exit(exit_code)


class AppStarting(Event):
    pass


class AppStarted(Event):
    pass


class AppStopping(Event):
    pass





from dataclasses import dataclass
from typing import Callable, Optional

import easy_pysy as ez
from easy_pysy import Interval


@dataclass
class Loop:
    interval_ms: int
    callback: Callable
    stop_app_on_error: bool
    auto_start: bool
    interval: Optional[Interval] = None

    def start(self):
        ez.require(not self.interval, "Loop is already running")
        self.interval = Interval(self.interval_ms, self.callback, self.on_error)
        self.interval.start()
        ez.emit(LoopStarted(loop=self))

    def stop(self):
        self.interval.cancel()
        self.interval = None
        ez.emit(LoopStopped(loop=self))

    def on_error(self, exception: BaseException):
        ez.exception(f'Loop execution failed: {exception}')
        if self.stop_app_on_error:
            ez.shutdown()

    @property
    def running(self):
        return self.interval is not None and self.interval.is_alive()


class LoopStarted(ez.Event):
    loop: Loop


class LoopStopped(ez.Event):
    loop: Loop


loops: list[Loop] = []


@ez.on(ez.AppStarting)
def start(event: ez.AppStarting):
    for loop in loops:
        if loop.auto_start:
            ez.debug(f'Starting {loop}')
            loop.start()


@ez.on(ez.AppStopping)
def stop(event: ez.AppStopping):
    for loop in loops:
        if loop.running:
            loop.stop()


def get_loop(callback: Callable) -> Optional[Loop]:
    # TODO: nice to have :
    # ez._(self.loops).find(lambda loop: loop.callback == callback)
    for loop in loops:
        if loop.callback == callback:
            return loop
    return None


def loop(every_ms: int, stop_app_on_error=True, auto_start=True):
    def decorator(func):
        loops.append(Loop(every_ms, func, stop_app_on_error, auto_start))
        return func
    return decorator



import logging
import time
from functools import wraps
from typing import Type

from easy_pysy.utils.common import require


logger = logging.getLogger(__name__)


def retry(times: int, exception_class: Type = BaseException, sleep=0):
    require(times >= 1, "times should be >= in @retry")

    def decorated(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempt = 1
            last_exception = None
            while attempt <= times:
                try:
                    return func(*args, **kwargs)
                except exception_class as exc:
                    logger.exception(f'Execution of {func} failed (attempt: {attempt}/{times})')
                    attempt += 1
                    last_exception = exc
                    if attempt <= times:
                        logger.debug(f'Will retry in {sleep} seconds')
                        time.sleep(sleep)
                    else:
                        raise exc
            raise last_exception
        return wrapper
    return decorated



import threading


class Singleton:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance
